@@ -48,12 +48,13 @@ import Data.Maybe
 import qualified Data.Set as Set
 import Data.Set (Set)
 
-import Language.Haskell.TH.Datatype
+import Language.Haskell.TH.Datatype ( freeVariables, freeVariablesWellScoped
+                                    , quantifyType, resolveTypeSynonyms )
 import Language.Haskell.TH.Datatype.TyVarBndr
 import Language.Haskell.TH.Instances ()
 import Language.Haskell.TH.Syntax hiding ( lift )
 
-import Language.Haskell.TH.Desugar.Util
+import Language.Haskell.TH.Desugar.Util as Util
 
 -- | Like @reify@ from Template Haskell, but looks also in any not-yet-typechecked
 -- declarations. To establish this list of not-yet-typechecked declarations,
@@ -93,26 +94,30 @@ reifyFail name =
 -- Utilities
 ---------------------------------
 
--- | Extract the @TyVarBndr@s and constructors given the @Name@ of a type
+-- | Extract the 'DataFlavor', 'TyVarBndr's and constructors given the 'Name'
+-- of a type.
 getDataD :: DsMonad q
          => String       -- ^ Print this out on failure
          -> Name         -- ^ Name of the datatype (@data@ or @newtype@) of interest
-         -> q ([TyVarBndrUnit], [Con])
+         -> q (DataFlavor, [TyVarBndrUnit], [Con])
 getDataD err name = do
   info <- reifyWithLocals name
   dec <- case info of
            TyConI dec -> return dec
            _ -> badDeclaration
   case dec of
-    DataD _cxt _name tvbs mk cons _derivings -> go tvbs mk cons
-    NewtypeD _cxt _name tvbs mk con _derivings -> go tvbs mk [con]
+    DataD _cxt _name tvbs mk cons _derivings -> go Data tvbs mk cons
+    NewtypeD _cxt _name tvbs mk con _derivings -> go Newtype tvbs mk [con]
+#if __GLASGOW_HASKELL__ >= 906
+    TypeDataD _name tvbs mk cons -> go Util.TypeData tvbs mk cons
+#endif
     _ -> badDeclaration
   where
-    go tvbs mk cons = do
+    go df tvbs mk cons = do
       let k = fromMaybe (ConT typeKindName) mk
       extra_tvbs <- mkExtraKindBinders k
       let all_tvbs = tvbs ++ extra_tvbs
-      return (all_tvbs, cons)
+      return (df, all_tvbs, cons)
 
     badDeclaration =
           fail $ "The name (" ++ (show name) ++ ") refers to something " ++
@@ -159,7 +164,7 @@ dataConNameToCon con_name = do
   -- we need to get the field ordering from the constructor. We must reify
   -- the constructor to get the tycon, and then reify the tycon to get the `Con`s
   type_name <- dataConNameToDataName con_name
-  (_, cons) <- getDataD "This seems to be an error in GHC." type_name
+  (_, _, cons) <- getDataD "This seems to be an error in GHC." type_name
   let m_con = List.find (any (con_name ==) . get_con_name) cons
   case m_con of
     Just con -> return con
@@ -263,6 +268,9 @@ reifyInDec n _    dec@(ClosedTypeFamilyD (TypeFamilyHead n' _ _ _) _) | n `nameM
 reifyInDec n decs (PatSynD n' _ _ _) | n `nameMatches` n'
   = Just (n', mkPatSynI n decs)
 #endif
+#if __GLASGOW_HASKELL__ >= 906
+reifyInDec n _ dec@(TypeDataD n' _ _ _) | n `nameMatches` n' = Just (n', TyConI dec)
+#endif
 
 reifyInDec n decs (DataD _ ty_name tvbs _mk cons _)
   | Just info <- maybeReifyCon n decs ty_name (map tvbToTANormalWithSig tvbs) cons
@@ -286,6 +294,11 @@ reifyInDec n decs (InstanceD _ _ _ sub_decs)
     reify_in_instance dec@(DataInstD {})    = reifyInDec n (sub_decs ++ decs) dec
     reify_in_instance dec@(NewtypeInstD {}) = reifyInDec n (sub_decs ++ decs) dec
     reify_in_instance _                     = Nothing
+#if __GLASGOW_HASKELL__ >= 801
+reifyInDec n decs (PatSynD pat_syn_name args _ _)
+  | Just (n', full_sel_ty) <- maybeReifyPatSynRecSelector n decs pat_syn_name args
+  = Just (n', VarI n full_sel_ty Nothing)
+#endif
 #if __GLASGOW_HASKELL__ >= 807
 reifyInDec n decs (DataInstD _ _ lhs _ cons _)
   | (ConT ty_name, tys) <- unfoldType lhs
@@ -303,6 +316,11 @@ reifyInDec n decs (NewtypeInstD _ ty_name tys _ con _)
   | Just info <- maybeReifyCon n decs ty_name (map TANormal tys) [con]
   = Just info
 #endif
+#if __GLASGOW_HASKELL__ >= 906
+reifyInDec n decs (TypeDataD ty_name tvbs _mk cons)
+  | Just info <- maybeReifyCon n decs ty_name (map tvbToTANormalWithSig tvbs) cons
+  = Just info
+#endif
 
 reifyInDec _ _ _ = Nothing
 
@@ -320,22 +338,181 @@ maybeReifyCon n _decs ty_name ty_args cons
       -- we don't try to ferret out naughty record selectors.
   = Just (n', VarI n full_sel_ty Nothing)
   where
-    extract_rec_sel_info :: RecSelInfo -> ([TyVarBndrUnit], Type, Type)
+    extract_rec_sel_info :: RecSelInfo -> ([TyVarBndrSpec], Type, Type)
       -- Returns ( Selector type variable binders
       --         , Record field type
       --         , constructor result type )
     extract_rec_sel_info rec_sel_info =
       case rec_sel_info of
-        RecSelH98 sel_ty -> (h98_tvbs, sel_ty, h98_res_ty)
-        RecSelGADT sel_ty con_res_ty ->
-          ( freeVariablesWellScoped [con_res_ty, sel_ty]
-          , sel_ty, con_res_ty)
+        RecSelH98 sel_ty ->
+          ( changeTVFlags SpecifiedSpec h98_tvbs
+          , sel_ty
+          , h98_res_ty
+          )
+        RecSelGADT mb_con_tvbs sel_ty con_res_ty ->
+          let -- If the GADT constructor type signature explicitly quantifies
+              -- its type variables, make sure to use that same order in the
+              -- record selector's type.
+              con_tvbs' =
+                case mb_con_tvbs of
+                  Just con_tvbs -> con_tvbs
+                  Nothing ->
+                    changeTVFlags SpecifiedSpec $
+                    freeVariablesWellScoped [con_res_ty, sel_ty] in
+          ( con_tvbs'
+          , sel_ty
+          , con_res_ty
+          )
 
-    h98_tvbs   = freeVariablesWellScoped $ map probablyWrongUnTypeArg ty_args
+    h98_tvbs   = freeVariablesWellScoped $
+                 map probablyWrongUnTypeArg ty_args
     h98_res_ty = applyType (ConT ty_name) ty_args
 
 maybeReifyCon _ _ _ _ _ = Nothing
 
+#if __GLASGOW_HASKELL__ >= 801
+-- | Attempt to reify the type of a pattern synonym record selector @n@.
+-- The algorithm for computing this type works as follows:
+--
+-- 1. Reify the type of the parent pattern synonym. Broadly speaking, this
+--    will look something like:
+--
+--    @
+--    pattern P :: forall <req_tvbs>. req_cxt =>
+--                 forall <prov_tvbs>. prov_cxt =>
+--                 arg_ty_1 -> ... -> arg_ty_k -> res
+--    @
+--
+-- 2. Check if @P@ is a record pattern synonym. If it isn't a record pattern
+--    synonym, return 'Nothing'. If it is a record pattern synonym, it will
+--    have @k@ record selectors @sel_1@, ..., @sel_k@.
+--
+-- 3. Check if @n@ is equal to some @sel_i@. If it isn't equal to any of them,
+--    return @Nothing@. If it is equal to some @sel_i@, then return 'Just'
+--    @sel_i@ paired with the following type:
+--
+--    @
+--    sel_i :: forall <req_tvbs>. req_cxt => res -> arg_ty_i
+--    @
+maybeReifyPatSynRecSelector ::
+  Name -> [Dec] -> Name -> PatSynArgs -> Maybe (Named Type)
+maybeReifyPatSynRecSelector n decs pat_syn_name pat_syn_args =
+  case pat_syn_args of
+    -- Part (2) in the Haddocks
+    RecordPatSyn fld_names
+      -> firstMatch match_pat_syn_rec_sel $
+         zip fld_names pat_syn_ty_vis_args
+    _ -> Nothing
+  where
+    -- Part (3) in the Haddocks
+    match_pat_syn_rec_sel :: (Name, Type) -> Maybe (Named Type)
+    match_pat_syn_rec_sel (n', field_ty)
+      | n `nameMatches` n'
+      = Just ( n'
+             , -- See Note [Use unSigType in maybeReifyCon]
+               unSigType $
+               maybeForallT pat_syn_ty_tvbs pat_syn_ty_req_cxt $
+               ArrowT `AppT` pat_syn_ty_res `AppT` field_ty
+             )
+    match_pat_syn_rec_sel _
+      = Nothing
+
+    -- The type of the pattern synonym to which this record selector belongs,
+    -- as described in part (1) in the Haddocks.
+    pat_syn_ty :: Type
+    pat_syn_ty =
+      case findPatSynType pat_syn_name decs of
+        Just ty -> ty
+        Nothing -> no_type n
+
+    pat_syn_ty_args :: FunArgs
+    pat_syn_ty_res :: Type
+    (pat_syn_ty_args, pat_syn_ty_res) =
+      unravelType pat_syn_ty
+
+    -- Decompose a pattern synonym type into the constituent parts described in
+    -- part (1) in the Haddocks. The Haddocks present an idealized form of
+    -- pattern synonym type signature where the required and provided foralls
+    -- and contexts are made explicit. In reality, some of these parts may be
+    -- omitted, so we have to be careful to handle every combination of
+    -- explicit and implicit parts.
+    pat_syn_ty_tvbs :: [TyVarBndrSpec]
+    pat_syn_ty_req_cxt :: Cxt
+    pat_syn_ty_vis_args :: [Type]
+    (pat_syn_ty_tvbs, pat_syn_ty_req_cxt, pat_syn_ty_vis_args) =
+      case pat_syn_ty_args of
+        -- Both the required foralls and context are explicit.
+        --
+        -- The provided foralls and context may be explicit or implicit, but it
+        -- doesn't really matter, as the type of a pattern synonym record
+        -- selector only cares about the required foralls and context.
+        -- Similarly for all cases below this one.
+        FAForalls (ForallInvis req_tvbs) (FACxt req_cxt args) ->
+          ( req_tvbs
+          , req_cxt
+          , mapMaybe vis_arg_anon_maybe $ filterVisFunArgs args
+          )
+
+        -- Only the required foralls are explicit. We can assume that there is
+        -- no required context due to the case above not matching.
+        FAForalls (ForallInvis req_tvbs) args ->
+          ( req_tvbs
+          , []
+          , mapMaybe vis_arg_anon_maybe $ filterVisFunArgs args
+          )
+
+        -- The required context is explicit, but the required foralls are
+        -- implicit. As a result, the order of type variables in the outer
+        -- forall in the type of the pattern synonym is determined by the usual
+        -- left-to-right scoped sort.
+        --
+        -- Note that there may be explicit, provided foralls in this case. For
+        -- example, consider this example:
+        --
+        -- @
+        -- data T a where
+        --   MkT :: b -> T (Maybe b)
+        --
+        -- pattern X :: Show a => forall b. (a ~ Maybe b) => b -> T a
+        -- pattern X{unX} = MkT unX
+        -- @
+        --
+        -- You might worry that the type of @unX@ would need to mention @b@.
+        -- But actually, you can't use @unX@ as a top-level record selector in
+        -- the first place! If you try to do so, GHC will throw the following
+        -- error:
+        --
+        -- @
+        -- Cannot use record selector `unX' as a function due to escaped type variables
+        -- @
+        --
+        -- As a result, we choose not to care about this corner case. We could
+        -- imagine trying to detect this sort of thing here and throwing a
+        -- similar error message, but detecting which type variables do or do
+        -- not escape is tricky in general. (See the Haddocks for
+        -- getRecordSelectors in L.H.TH.Desugar for more on this point.) As a
+        -- result, we don't even bother trying. Similarly for the case below.
+        FACxt req_cxt args ->
+          ( changeTVFlags SpecifiedSpec $
+            freeVariablesWellScoped [pat_syn_ty]
+          , req_cxt
+          , mapMaybe vis_arg_anon_maybe $ filterVisFunArgs args
+          )
+
+        -- The required foralls are implicit. We can assume that there is no
+        -- required context due to the case above not matching.
+        args ->
+          ( changeTVFlags SpecifiedSpec $
+            freeVariablesWellScoped [pat_syn_ty]
+          , []
+          , mapMaybe vis_arg_anon_maybe $ filterVisFunArgs args
+          )
+
+vis_arg_anon_maybe :: VisFunArg -> Maybe Type
+vis_arg_anon_maybe (VisFAAnon ty) = Just ty
+vis_arg_anon_maybe (VisFADep{})   = Nothing
+#endif
+
 {-
 Note [Use unSigType in maybeReifyCon]
 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
@@ -361,7 +538,9 @@ con_to_type :: [TyVarBndrUnit] -- The type variables bound by a data type head.
 con_to_type h98_tvbs h98_result_ty con =
   case go con of
     (is_gadt, ty) | is_gadt   -> ty
-                  | otherwise -> maybeForallT h98_tvbs [] ty
+                  | otherwise -> maybeForallT
+                                   (changeTVFlags SpecifiedSpec h98_tvbs)
+                                   [] ty
   where
     -- Note that we deliberately ignore linear types and use (->) everywhere.
     -- See [Gracefully handling linear types] in L.H.TH.Desugar.Core.
@@ -597,13 +776,11 @@ mkArrows :: [Type] -> Type -> Type
 mkArrows []     res_ty = res_ty
 mkArrows (t:ts) res_ty = AppT (AppT ArrowT t) $ mkArrows ts res_ty
 
-maybeForallT :: [TyVarBndrUnit] -> Cxt -> Type -> Type
+maybeForallT :: [TyVarBndrSpec] -> Cxt -> Type -> Type
 maybeForallT tvbs cxt ty
   | null tvbs && null cxt        = ty
-  | ForallT tvbs2 cxt2 ty2 <- ty = ForallT (tvbs_spec ++ tvbs2) (cxt ++ cxt2) ty2
-  | otherwise                    = ForallT tvbs_spec cxt ty
-  where
-    tvbs_spec = changeTVFlags SpecifiedSpec tvbs
+  | ForallT tvbs2 cxt2 ty2 <- ty = ForallT (tvbs ++ tvbs2) (cxt ++ cxt2) ty2
+  | otherwise                    = ForallT tvbs cxt ty
 
 findCon :: Name -> [Con] -> Maybe (Named Con)
 findCon n = firstMatch match_con
@@ -628,24 +805,65 @@ findCon n = firstMatch match_con
 
 data RecSelInfo
   = RecSelH98  Type -- The record field's type
-  | RecSelGADT Type -- The record field's type
+  | RecSelGADT (Maybe [TyVarBndrSpec])
+                    -- If the data constructor explicitly quantifies its type
+                    -- variables with a forall, this will be Just. Otherwise,
+                    -- this will be Nothing.
+               Type -- The record field's type
                Type -- The GADT return type
 
 findRecSelector :: Name -> [Con] -> Maybe (Named RecSelInfo)
-findRecSelector n = firstMatch match_con
+findRecSelector n = firstMatch (match_con Nothing)
   where
-    match_con :: Con -> Maybe (Named RecSelInfo)
-    match_con (RecC _ vstys)            = fmap (liftSnd RecSelH98) $
-                                          firstMatch match_rec_sel vstys
-    match_con (RecGadtC _ vstys ret_ty) = fmap (liftSnd (`RecSelGADT` ret_ty)) $
-                                          firstMatch match_rec_sel vstys
-    match_con (ForallC _ _ c)           = match_con c
-    match_con _                         = Nothing
+    match_con :: Maybe [TyVarBndrSpec] -> Con -> Maybe (Named RecSelInfo)
+    match_con mb_tvbs con =
+      case con of
+        RecC _ vstys ->
+          fmap (liftSnd RecSelH98) $
+          firstMatch match_rec_sel vstys
+        RecGadtC _ vstys ret_ty ->
+          fmap (liftSnd (\field_ty ->
+            RecSelGADT (fmap (filter_ret_tvs ret_ty) mb_tvbs) field_ty ret_ty)) $
+          firstMatch match_rec_sel vstys
+        ForallC tvbs _ c ->
+          -- This is the only recursive case, and it is also the place where
+          -- the type variable binders are determined (hence the use of Just
+          -- below). Note that GHC forbids nested foralls in GADT constructor
+          -- type signatures, so it is guaranteed that if a type variable in
+          -- the rest of the type signature appears free, then its binding site
+          -- can be found in one of these binders found in this case.
+          match_con (Just tvbs) c
+        _ -> Nothing
 
     match_rec_sel (n', _, sel_ty)
       | n `nameMatches` n' = Just (n', sel_ty)
     match_rec_sel _        = Nothing
 
+    -- There may be type variables in the type of a GADT constructor that do
+    -- not appear in the type of a record selector. For example, consider:
+    --
+    --   data G a where
+    --     MkG :: forall a b. { x :: a, y :: b } -> G a
+    --
+    -- The type of `x` will only quantify `a` and not `b`:
+    --
+    --   x :: forall a. G a -> a
+    --
+    -- Accordingly, we must filter out any type variables in the GADT
+    -- constructor type that do not appear free in the return type. Note that
+    -- this implies that we cannot support reifying the type of `y`, as `b`
+    -- does not appear free in `G a`. This does not bother us, however, as we
+    -- make no attempt to support naughty record selectors. (See the Haddocks
+    -- for getRecordSelectors in L.H.TH.Desugar for more on this point.)
+    --
+    -- This mirrors the implementation of mkOneRecordSelector in GHC:
+    -- https://gitlab.haskell.org/ghc/ghc/-/blob/37cfe3c0f4fb16189bbe3bb735f758cd6e3d9157/compiler/GHC/Tc/TyCl/Utils.hs#L908-909
+    filter_ret_tvs :: Type -> [TyVarBndrSpec] -> [TyVarBndrSpec]
+    filter_ret_tvs ret_ty =
+      filter (\tvb -> tvName tvb `Set.member` ret_fvs)
+      where
+        ret_fvs = Set.fromList $ freeVariables [ret_ty]
+
 ---------------------------------
 -- Reifying fixities
 ---------------------------------
@@ -802,6 +1020,11 @@ match_cusk n (ClassD _ n' tvbs _ sub_decs)
                                         , Just tvb_kind <- [tvb_kind_maybe tvb]
                                         ]
   = firstMatch (find_assoc_type_kind n cls_tvb_kind_map) sub_decs
+#if __GLASGOW_HASKELL__ >= 906
+match_cusk n (TypeDataD n' tvbs m_ki _)
+  | n `nameMatches` n'
+  = datatype_kind tvbs m_ki
+#endif
 match_cusk _ _ = Nothing
 
 -- Uncover the kind of an associated type family. There is an invariant
@@ -993,7 +1216,7 @@ lookupValueNameWithLocals = lookupNameWithLocals False
 lookupTypeNameWithLocals :: DsMonad q => String -> q (Maybe Name)
 lookupTypeNameWithLocals = lookupNameWithLocals True
 
-lookupNameWithLocals :: DsMonad q => Bool -> String -> q (Maybe Name)
+lookupNameWithLocals :: forall q. DsMonad q => Bool -> String -> q (Maybe Name)
 lookupNameWithLocals ns s = do
     mb_name <- qLookupName ns s
     case mb_name of
@@ -1006,20 +1229,22 @@ lookupNameWithLocals ns s = do
       decs <- localDeclarations
       let mb_infos = map (reifyInDec built_name decs) decs
           infos = catMaybes mb_infos
-      return $ firstMatch (if ns then find_type_name
-                                 else find_value_name) infos
+      firstMatchM (if ns then find_type_name
+                         else find_value_name) infos
 
     -- These functions work over Named Infos so we can avoid performing
     -- tiresome pattern-matching to retrieve the name associated with each Info.
-    find_type_name, find_value_name :: Named Info -> Maybe Name
-    find_type_name (n, info) =
-      case infoNameSpace info of
+    find_type_name, find_value_name :: Named Info -> q (Maybe Name)
+    find_type_name (n, info) = do
+      name_space <- lookupInfoNameSpace info
+      pure $ case name_space of
         TcClsName -> Just n
         VarName   -> Nothing
         DataName  -> Nothing
 
-    find_value_name (n, info) =
-      case infoNameSpace info of
+    find_value_name (n, info) = do
+      name_space <- lookupInfoNameSpace info
+      pure $ case name_space of
         VarName   -> Just n
         DataName  -> Just n
         TcClsName -> Nothing
@@ -1060,22 +1285,33 @@ reifyNameSpace n@(Name _ nf) =
     -- For other names, we must use reification to determine what NameSpace
     -- it lives in (if any).
     _ -> do mb_info <- reifyWithLocals_maybe n
-            pure $ fmap infoNameSpace mb_info
+            traverse lookupInfoNameSpace mb_info
 
--- | Determine a name's 'NameSpace' from its 'Info'.
-infoNameSpace :: Info -> NameSpace
-infoNameSpace info =
+-- | Look up a name's 'NameSpace' from its 'Info'.
+lookupInfoNameSpace :: DsMonad q => Info -> q NameSpace
+lookupInfoNameSpace info =
   case info of
-    ClassI{}     -> TcClsName
-    TyConI{}     -> TcClsName
-    FamilyI{}    -> TcClsName
-    PrimTyConI{} -> TcClsName
-    TyVarI{}     -> TcClsName
+    ClassI{}     -> pure TcClsName
+    TyConI{}     -> pure TcClsName
+    FamilyI{}    -> pure TcClsName
+    PrimTyConI{} -> pure TcClsName
+    TyVarI{}     -> pure TcClsName
 
-    ClassOpI{}   -> VarName
-    VarI{}       -> VarName
+    ClassOpI{}   -> pure VarName
+    VarI{}       -> pure VarName
 
-    DataConI{}   -> DataName
+    DataConI _dc_name _dc_ty parent_name -> do
+      -- DataConI usually refers to a value-level Name, but it could also refer
+      -- to a type-level 'Name' if the data constructor corresponds to a
+      -- @type data@ declaration. In order to know for sure, we must perform
+      -- some additional reification.
+      mb_parent_info <- reifyWithLocals_maybe parent_name
+      pure $ case mb_parent_info of
+#if __GLASGOW_HASKELL__ >= 906
+        Just (TyConI (TypeDataD {}))
+          -> TcClsName
+#endif
+        _ -> DataName
 #if __GLASGOW_HASKELL__ >= 801
-    PatSynI{}    -> DataName
+    PatSynI{}    -> pure DataName
 #endif
