@@ -6,9 +6,9 @@ rae@cs.brynmawr.edu
 Utility functions for th-desugar package.
 -}
 
-{-# LANGUAGE CPP, DeriveDataTypeable, RankNTypes, ScopedTypeVariables,
-             TupleSections, AllowAmbiguousTypes, TemplateHaskellQuotes,
-             TypeApplications #-}
+{-# LANGUAGE CPP, DeriveDataTypeable, DeriveGeneric, DeriveLift, RankNTypes,
+             ScopedTypeVariables, TupleSections, AllowAmbiguousTypes,
+             TemplateHaskellQuotes, TypeApplications #-}
 
 module Language.Haskell.TH.Desugar.Util (
   newUniqueName,
@@ -20,7 +20,7 @@ module Language.Haskell.TH.Desugar.Util (
   thirdOf3, splitAtList, extractBoundNamesDec,
   extractBoundNamesPat,
   tvbToType, tvbToTypeWithSig, tvbToTANormalWithSig,
-  nameMatches, thdOf3, liftFst, liftSnd, firstMatch,
+  nameMatches, thdOf3, liftFst, liftSnd, firstMatch, firstMatchM,
   unboxedSumDegree_maybe, unboxedSumNameDegree_maybe,
   tupleDegree_maybe, tupleNameDegree_maybe, unboxedTupleDegree_maybe,
   unboxedTupleNameDegree_maybe, splitTuple_maybe,
@@ -29,7 +29,8 @@ module Language.Haskell.TH.Desugar.Util (
   unSigType, unfoldType, ForallTelescope(..), FunArgs(..), VisFunArg(..),
   filterVisFunArgs, ravelType, unravelType,
   TypeArg(..), applyType, filterTANormals, probablyWrongUnTypeArg,
-  bindIP
+  bindIP,
+  DataFlavor(..)
   ) where
 
 import Prelude hiding (mapM, foldl, concatMap, any)
@@ -43,10 +44,11 @@ import Language.Haskell.TH.Syntax
 import qualified Control.Monad.Fail as Fail
 import Data.Foldable
 import qualified Data.Kind as Kind
-import Data.Generics hiding ( Fixity )
+import Data.Generics ( Data, Typeable, everything, extM, gmapM, mkQ )
 import Data.Traversable
 import Data.Maybe
 import GHC.Classes ( IP )
+import GHC.Generics ( Generic )
 import Unsafe.Coerce ( unsafeCoerce )
 
 ----------------------------------------
@@ -492,6 +494,9 @@ expectJustM err Nothing  = Fail.fail err
 firstMatch :: (a -> Maybe b) -> [a] -> Maybe b
 firstMatch f xs = listToMaybe $ mapMaybe f xs
 
+firstMatchM :: Monad m => (a -> m (Maybe b)) -> [a] -> m (Maybe b)
+firstMatchM f xs = listToMaybe <$> mapMaybeM f xs
+
 -- | Semi-shallow version of 'everywhereM' - does not recurse into children of nodes of type @a@ (only applies the handler to them).
 --
 -- >>> topEverywhereM (pure . fmap (*10) :: [Integer] -> Identity [Integer]) ([1,2,3] :: [Integer], "foo" :: String)
@@ -532,3 +537,11 @@ starKindName = ''(Kind.*)
 uniStarKindName :: Name
 uniStarKindName = ''(Kind.★)
 #endif
+
+-- | Is a data type or data instance declaration a @newtype@ declaration, a
+-- @data@ declaration, or a @type data@ declaration?
+data DataFlavor
+  = Newtype  -- ^ @newtype@
+  | Data     -- ^ @data@
+  | TypeData -- ^ @type data@
+  deriving (Eq, Show, Data, Generic, Lift)
